
# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/python/arcor2_mocks:arcor2_mocks_dist

from setuptools import setup

setup(**{
    'author': 'Robo@FIT',
    'author_email': 'imaterna@fit.vut.cz',
    'classifiers': [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3.8',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering',
    ],
    'description': 'ARCOR2 integration for Kinali services.',
    'license': 'LGPL',
    'long_description': """# arcor2_mocks
# Changelog

The format is based on [Keep a Changelog](https://keepachangelog.com/en/1.0.0/),


## [0.8.0] - WIP
### Changed
- Mocks updated to match Project 0.4.0 and Scene 0.2.0.""",
    'long_description_content_type': 'text/markdown',
    'name': 'arcor2_mocks',
    'version': '0.8.0',
    'package_dir': {
        '': 'src',
    },
    'packages': (
        'arcor2_mocks',
        'arcor2_mocks.scripts',
    ),
    'namespace_packages': (
    ),
    'package_data': {
        'arcor2_mocks': (
            'VERSION',
            'py.typed',
        ),
    },
    'install_requires': (
        'Flask==1.1.2',
        'PyYAML==5.3.1',
        'apispec-webframeworks==0.5.2',
        'apispec==3.3.2',
        'arcor2==0.8.0',
        'dataclasses-jsonschema-rereleased[fast-validation]==2.12.2',
        'flask-swagger-ui==3.25.0',
        'flask_cors==3.0.8',
        'pyhumps==1.6.1',
    ),
    'entry_points': {
        'console_scripts': [
            'arcor2_mock_project=arcor2_mocks.scripts.project',
            'arcor2_mock_scene=arcor2_mocks.scripts.scene',
        ],
    },
})
